from pydantic import BaseModel
from typing import Optional, Literal

class CreateNode(BaseModel):
    human_rep: str
    lean_rep: str 
    verification: Optional[int] 

class CreateStatement(CreateNode):
    category: str

class CreateImplication(CreateNode):
    logic_op: Literal['AND', 'OR']
    premises_ids: list[str]
    concludes_ids: list[str]

class NodePatchRequest(BaseModel):
    human_rep: Optional[str] = None
    lean_rep: Optional[str] = None
    verification: Optional[int] = None



from sqlmodel import SQLModel, Field, Relationship, Column
from typing import Optional, Dict, Any, TYPE_CHECKING
from datetime import datetime
from sqlalchemy.types import JSON

if TYPE_CHECKING:
    from app.models.auth import Agent

class NodeWorkBase(SQLModel):
    """Common fields for any work item (Patch, Comment, etc)"""
    target_node_id: str = Field(index=True)
    agent_id: str = Field(index=True, foreign_key="agent.id")
    created_at: datetime = Field(default_factory=datetime.utcnow)

class NodePatchBase(NodeWorkBase):
    update_data: Dict[str, Any] = Field(sa_column=Column(JSON))

class NodePatch(NodePatchBase, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    agent: "Agent" = Relationship(back_populates="node_patches")

class NodePatchRead(NodePatchBase):
    id: int 

class NodeCommentBase(NodeWorkBase):
    comment: str

class NodeComment(NodeCommentBase, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    agent: "Agent" = Relationship(back_populates="node_comments")

class NodeCommentRead(NodeCommentBase):
    id: int


# neomodel graph nodes

from neomodel import (
    StructuredNode, 
    StringProperty, 
    JSONProperty, 
    DateTimeProperty, 
    RelationshipTo, 
    RelationshipFrom,
    UniqueIdProperty,
    IntegerProperty
)
from datetime import datetime
from enum import IntEnum

class VerificationLevel(IntEnum):
    REJECTED = 0          # Proven false
    SPECULATIVE = 1       # LLM/Human guess
    NUMERICAL = 2         # Verified in Python but not Lean
    FORMAL_SKETCH = 3     # Lean code exists, assumes 'sorry'
    VERIFIED = 4          # Compiled successfully

VERIFICATION_CHOICES = [(status.value, status.name) for status in VerificationLevel]

class PolymathBase(StructuredNode):
    """
    Abstract base class for all nodes in the protocol.
    """
    __abstract_node__ = True

    # Unique ID 
    uid = UniqueIdProperty()
    
    # Metadata
    created_at = DateTimeProperty(default_datetime=datetime.utcnow)
    updated_at = DateTimeProperty(default_datetime=datetime.utcnow)
    
    # Who created this node (Agent ID)
    author_id = StringProperty(required=True)
    
    # 2 dialects: human or lean
    human_rep = StringProperty()
    lean_rep = StringProperty()
    
    # Verification Metadata
    verification = IntegerProperty(
        choices=VERIFICATION_CHOICES, 
        default=VerificationLevel.SPECULATIVE
    )


class Statement(PolymathBase):
    """
    Represents a Theorem, Axiom, Lemma, or Definition.
    The 'Dot' in the graph.
    """
    # theorem, lemma, axiom, or definition
    category = StringProperty(default="CONJECTURE")
    
    proven_by = RelationshipFrom('Implication', 'IS_PROOF')
    
    supports = RelationshipTo('Implication', 'IS_PREMISE')

class Implication(PolymathBase):
    """
    Represents the logical step or proof.
    The 'Hyperedge' in the graph (Reified as a Node).
    """
    # and, or
    logic_operator = StringProperty(default="AND")
    
    premises = RelationshipFrom('Statement', 'IS_PREMISE')
    
    concludes = RelationshipTo('Statement', 'IS_PROOF')



from pydantic import BaseModel
from typing import Literal, List, Union
from datetime import datetime
from polymath_schemas.graph import VerificationLevel

class PolymathBaseRead(BaseModel):
    uid: str
    author_id: str
    created_at: datetime
    updated_at: datetime
    human_rep: str | None = None
    lean_rep: str | None = None
    verification: VerificationLevel

class StatementRead(PolymathBaseRead):
    node_type: Literal["Statement"] = "Statement"  # Discriminator
    category: str

class ImplicationRead(PolymathBaseRead):
    node_type: Literal["Implication"] = "Implication" # Discriminator
    logic_operator: str

class UnifiedNodeResponse(BaseModel):
    node_data: Union[StatementRead, ImplicationRead] 
    patches: List[NodePatchRead]
    comments: List[NodeCommentRead]


from sqlmodel import SQLModel, Field, Relationship
from typing import Optional, TYPE_CHECKING
from app.models.graph import VerificationLevel
from typing import List

if TYPE_CHECKING:
    from app.models.node_work import NodePatch, NodeComment

class Role(SQLModel, table=True):
    id: Optional[int] = Field(default=None, primary_key=True)
    
    name: str = Field(index=True, unique=True) 
    
    # This integer determines the ceiling of what this role can commit
    highest_verification_allowed: int = Field(
        default=VerificationLevel.SPECULATIVE,
        description="The maximum VerificationLevel int value this role is allowed to assign."
    )

    agents: List["Agent"] = Relationship(back_populates="role")


class Agent(SQLModel, table=True):
    
    id: Optional[int] = Field(default=None, primary_key=True)

    name: str = Field(index=True)

    api_key_hash: str = Field(index=True, unique=True)

    role_id: Optional[int] = Field(default=None, foreign_key="role.id")
    
    role: Optional[Role] = Relationship(back_populates="agents")

    node_patches: List["NodePatch"] = Relationship(back_populates="agent")

    node_comments: List["NodeComment"] = Relationship(back_populates="agent")
